package com.menagerie.bakers

import androidx.compose.animation.ExperimentalSharedTransitionApi
import androidx.compose.animation.SharedTransitionLayout
import androidx.compose.foundation.isSystemInDarkTheme
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.Spacer
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.width
import androidx.compose.foundation.lazy.grid.GridCells
import androidx.compose.foundation.lazy.grid.LazyVerticalGrid
import androidx.compose.foundation.lazy.grid.rememberLazyGridState
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.automirrored.filled.ArrowBack
import androidx.compose.material.icons.automirrored.filled.ExitToApp
import androidx.compose.material.icons.filled.ExitToApp
import androidx.compose.material.icons.filled.Share
import androidx.compose.material.icons.filled.Settings
import androidx.compose.material3.Button
import androidx.compose.material3.Icon
import androidx.compose.material3.IconButton
import androidx.compose.material3.Scaffold
import androidx.compose.material3.Text
import androidx.compose.runtime.Composable
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateListOf
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.setValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.unit.dp
import androidx.navigation.NavType
import androidx.navigation.compose.NavHost
import androidx.navigation.compose.composable
import androidx.navigation.compose.rememberNavController
import androidx.navigation.navArgument
import bakersmenagerie.composeapp.generated.resources.Res
import bakersmenagerie.composeapp.generated.resources.auto_mode
import bakersmenagerie.composeapp.generated.resources.dark_mode
import bakersmenagerie.composeapp.generated.resources.light_mode
import bakersmenagerie.composeapp.generated.resources.shuffle
import bakersmenagerie.composeapp.generated.resources.tune
import com.menagerie.bakers.SettingsKeys.ANIM_KEY
import com.menagerie.bakers.SettingsKeys.AON_KEY
import com.menagerie.bakers.SettingsKeys.DESC_KEY
import com.menagerie.bakers.SettingsKeys.SORT_BY_KEY
import com.menagerie.bakers.SettingsKeys.THEME_KEY
import com.menagerie.bakers.SettingsKeys.TTT_KEY
import com.menagerie.bakers.ui.theme.MainTheme
import com.menagerie.bakers.view.details.RecipeDetails
import org.jetbrains.compose.resources.painterResource
import com.menagerie.bakers.view.list.RecipesListScreen
import com.menagerie.bakers.view.BookShelf
import com.menagerie.bakers.view.util.FilterCard
import com.menagerie.bakers.view.HomeScreen
import com.menagerie.bakers.view.util.InputFieldState
import com.menagerie.bakers.view.util.getFilteredRecipeList
import com.menagerie.bakers.view.util.getRecipeList
import com.menagerie.bakers.model.Recipe
import com.menagerie.bakers.model.TTT
import com.menagerie.bakers.model.TagType
import com.menagerie.bakers.view.SettingsMenu
import com.russhwolf.settings.Settings


enum class Theme {
    Auto,
    Light,
    Dark
}

enum class SortBy {
    Name,
    Time,
    Ingredients,
}

enum class RecipeAppScreen {
    List, Details,
}

@OptIn(ExperimentalSharedTransitionApi::class)
@Composable
fun App(
    onClose: () -> Unit = {}
) {

    //Grab an instance of the Settings object
    val settings by remember { mutableStateOf(Settings()) }

    //used to navigate through the different compose screens
    val navController = rememberNavController()

    //Grab and set Theme from Settings
    val themeSetting by remember { mutableStateOf(settings.getStringOrNull(THEME_KEY)) }
    var theme by remember {
        mutableStateOf(

            when (themeSetting) {
                Theme.Dark.name -> Theme.Dark
                Theme.Light.name -> Theme.Light
                else -> Theme.Auto
            }
        )
    }

    val isDarkTheme: Boolean? = when (theme) {
        Theme.Dark -> true
        Theme.Light -> false
        Theme.Auto -> null
    }

    val onThemeToggle = {
        theme = when (theme) {
            Theme.Auto -> Theme.Light
            Theme.Light -> Theme.Dark
            Theme.Dark -> Theme.Auto
        }

        when (theme) {
            Theme.Auto -> settings.putString(THEME_KEY, Theme.Auto.name)
            Theme.Light -> settings.putString(THEME_KEY, Theme.Light.name)
            Theme.Dark -> settings.putString(THEME_KEY, Theme.Dark.name)
        }
    }
    // END Theme


    // Grab and set Sort Order from Settings
    val sortBySetting by remember { mutableStateOf(settings.getStringOrNull(SORT_BY_KEY)) }
    var sortBy by remember {
        mutableStateOf(
            when (sortBySetting) {

                SortBy.Name.name -> SortBy.Name
                SortBy.Time.name -> SortBy.Time
                SortBy.Ingredients.name -> SortBy.Ingredients
                else -> SortBy.Name
            }
        )
    }

    val onSortToggle = {
        sortBy = when (sortBy) {
            SortBy.Name -> SortBy.Time
            SortBy.Time -> SortBy.Ingredients
            SortBy.Ingredients -> SortBy.Name
        }

        when (sortBy) {
            SortBy.Name -> settings.putString(SORT_BY_KEY, SortBy.Name.name)
            SortBy.Time -> settings.putString(SORT_BY_KEY, SortBy.Time.name)
            SortBy.Ingredients -> settings.putString(SORT_BY_KEY, SortBy.Ingredients.name)
        }
    }
    // END SortBy


    //Grab and set TTT Level from Settings
    val tttSetting by remember { mutableStateOf(settings.getString(TTT_KEY, TTT.TESTED.name)) }
    var tttBy by remember {
        mutableStateOf(
            when (tttSetting) {
                TTT.TRIED.name -> TTT.TRIED
                TTT.TESTED.name -> TTT.TESTED
                TTT.TRUE.name -> TTT.TRUE
                else -> TTT.TESTED
            }
        )
    }

    val onTTTToggle = {
        tttBy = when (tttBy) {
            TTT.TRIED -> TTT.TESTED
            TTT.TESTED -> TTT.TRUE
            TTT.TRUE -> TTT.TRIED
        }

        when (tttBy) {
            TTT.TRIED -> settings.putString(TTT_KEY, TTT.TRIED.name)
            TTT.TESTED -> settings.putString(TTT_KEY, TTT.TESTED.name)
            TTT.TRUE -> settings.putString(TTT_KEY, TTT.TRUE.name)
        }
    }
    // END TTT

    //Grab and set And/Or/Not from Settings
    val returnSetting by remember { mutableStateOf(settings.getBoolean(AON_KEY, false)) }
    var returnAnyMatch by remember { mutableStateOf(returnSetting) }
    //END AON

    //Grab and set Display Order from Settings
    val descSetting by remember { mutableStateOf(settings.getBoolean(DESC_KEY, false)) }
    var descending by remember { mutableStateOf(descSetting) }
    //END Display Order

    //Grab and set Animation State from Settings
    val animSetting by remember { mutableStateOf(settings.getBoolean(ANIM_KEY, true)) }
    var animate by remember { mutableStateOf(animSetting) }

    MainTheme(useDarkTheme = isDarkTheme ?: isSystemInDarkTheme()) {


        var showFilter by remember { mutableStateOf(false) }
        var searchBar by remember { mutableStateOf(false) }
        var search by remember { mutableStateOf("") }
        val tags = remember { mutableStateListOf<String>() }
        val recipeTags by remember { mutableStateOf(mutableMapOf<String, TagType>()) }
        var book by remember { mutableStateOf("") }
        val remove = remember { mutableStateListOf<String>() }
        var helpUs by remember { mutableStateOf(false) }

        val filteredItems = getFilteredRecipeList(
            tags = tags,
            search = search,
            lockTag = book,
            sortBy = sortBy,
            tttBy = tttBy,
            showMissing = helpUs,
            returnAny = returnAnyMatch,
            reverse = descending,
        )

        recipeTags.clear()

        for (recipe in filteredItems) {
            for (tag in recipe.tags) {
                if (!recipeTags.contains(tag.key))
                    recipeTags[tag.key] = tag.value
            }
        }

        if (showFilter) {
            FilterCard(
                recipeTags = recipeTags,
                activeTags = tags,
            ) {
                if (it.isEmpty()) {
                    tags.clear()
                }

                for (item in it)
                    if (tags.contains(item).not())
                        tags.add(item)

                remove.clear()

                for (item in tags)
                    if (it.contains(item).not())
                        remove.add(item)

                tags.removeAll(remove)

                showFilter = false
            }
        }


        val recipeCount = getRecipeList().size
        var currentRecipe: Recipe

        if (filteredItems.isNotEmpty())
            currentRecipe = filteredItems.first()

        Scaffold {
            Column {
                when (DisplayManager.device) {
                    Device.Android ->
                        if (searchBar)
                            Column {
                                Row {
                                    SettingsMenu(
                                        andOr = returnAnyMatch,
                                        theme = theme,
                                        sortBy = sortBy,
                                        ttt = tttBy,
                                        helpUs = helpUs,
                                        descending = descending,
                                        animate = animate,
                                        onAndOr = {
                                            returnAnyMatch = !returnAnyMatch
                                            settings.putBoolean(AON_KEY, returnAnyMatch)
                                        },
                                        onSortToggle = { onSortToggle.invoke() },
                                        onDesc = {
                                            descending = !descending
                                            settings.putBoolean(DESC_KEY, descending)
                                        },
                                        onAnim = {
                                            animate = !animate
                                            settings.putBoolean(ANIM_KEY, animate)
                                        },
                                        onTheme = { onThemeToggle.invoke() },
                                        onHelpToggle = { helpUs = !helpUs },
                                        onTTTToggle = { onTTTToggle.invoke() },
                                        onRandom = {
                                            currentRecipe = filteredItems.random()
                                            navController.navigate(
                                                RecipeAppScreen.Details.name.plus(
                                                    "/${currentRecipe.title}"
                                                )
                                            )
                                        },
                                        onBack = {},
                                        onClear = {settings.clear()}
                                    )

                                    Spacer(modifier = Modifier.weight(1f))

                                    InputFieldState(
                                        value = search,
                                        label = "Search ".plus(book),
                                        modifier = Modifier
                                            .width(if (DisplayManager.size == Size.Large || DisplayManager.orientation == Orientation.Landscape) 400.dp else 250.dp)
                                            .align(Alignment.CenterVertically)
                                    ) {
                                        search = it
                                    }

                                    if (DisplayManager.orientation == Orientation.Portrait)
                                        Spacer(modifier = Modifier.weight(1f))

                                    IconButton(
                                        onClick = { showFilter = true },
                                    ) {
                                        Icon(
                                            painter = painterResource(Res.drawable.tune), "Filter"
                                        )
                                    }
                                }
                                val listState = rememberLazyGridState()
                                LazyVerticalGrid(
                                    state = listState,
                                    columns = GridCells.Adaptive(minSize = 125.dp)
                                )
                                {
                                    items(tags.size) {
                                        Button(onClick = { tags.remove(tags[it]) }) {
                                            Text(text = tags[it])
                                        }
                                    }
                                }
                            }

                    Device.Desktop ->
                        if (searchBar)
                            Column {
                                Row {
                                    IconButton(
                                        onClick = { navController.popBackStack() },
                                        content = {
                                            Icon(
                                                imageVector = Icons.AutoMirrored.Filled.ArrowBack,
                                                contentDescription = "Back"
                                            )
                                        }
                                    )
                                    IconButton(
                                        onClick = { navController.navigate("Settings") },
                                        content = {
                                            Icon(
                                                imageVector = Icons.Default.Settings,
                                                contentDescription = ""
                                            )
                                        }
                                    )
                                    IconButton(
                                        onClick = {onThemeToggle.invoke()},
                                        content = {
                                            Icon(
                                                painter = painterResource(
                                                    when(theme) {
                                                        Theme.Auto -> Res.drawable.auto_mode
                                                        Theme.Light -> Res.drawable.light_mode
                                                        Theme.Dark -> Res.drawable.dark_mode
                                                    }
                                                ),
                                                contentDescription = "theme"
                                            )
                                        }
                                    )
                                    IconButton(
                                        onClick = {
                                            currentRecipe = filteredItems.random()
                                            navController.navigate(
                                                RecipeAppScreen.Details.name.plus(
                                                    "/${currentRecipe.title}"
                                                )
                                            )
                                        },
                                        content = {
                                            Icon(
                                                painter = painterResource(Res.drawable.shuffle),
                                                contentDescription = "random"
                                            )
                                        }
                                    )

                                    Spacer(modifier = Modifier.weight(1f))

                                    InputFieldState(
                                        value = search,
                                        label = "Search ".plus(book),
                                        modifier = Modifier
                                            .width(if (DisplayManager.size == Size.Large || DisplayManager.orientation == Orientation.Landscape) 400.dp else 250.dp)
                                            .align(Alignment.CenterVertically)
                                    ) {
                                        search = it
                                    }

                                    if (DisplayManager.orientation == Orientation.Portrait)
                                        Spacer(modifier = Modifier.weight(1f))

                                    IconButton(
                                        onClick = { showFilter = true },
                                        modifier = Modifier.align(Alignment.CenterVertically).padding(bottom = 8.dp)
                                    ) {
                                        Icon(
                                            painter = painterResource(Res.drawable.tune), "Filter",
                                        )
                                    }
                                    IconButton(
                                        onClick = onClose,
                                        modifier = Modifier.align(Alignment.CenterVertically).padding(bottom = 8.dp)
                                    ) {
                                        Icon(
                                            imageVector = Icons.AutoMirrored.Filled.ExitToApp, "Close",
                                        )
                                    }
                                }
                                Row {
                                    Spacer(modifier = Modifier.weight(1f))
                                    if (DisplayManager.orientation == Orientation.Landscape && (tags.isNotEmpty() || search.isNotEmpty() || book.isNotEmpty()))
                                        Text(
                                            text = "Showing ".plus(filteredItems.size)
                                                .plus(" out of ")
                                                .plus(recipeCount).plus(" recipes"),
                                        )
                                }
                                val listState = rememberLazyGridState()
                                LazyVerticalGrid(
                                    state = listState,
                                    columns = GridCells.Adaptive(minSize = 125.dp)
                                )
                                {
                                    items(tags.size) {
                                        Button(onClick = { tags.remove(tags[it]) }) {
                                            Text(text = tags[it])
                                        }
                                    }
                                }
                            }

                    Device.Web ->
                        if(searchBar)
                        Column {
                            Row {
                                IconButton(
                                    onClick = { navController.popBackStack() },
                                    content = {
                                        Icon(
                                            imageVector = Icons.AutoMirrored.Filled.ArrowBack,
                                            contentDescription = "Back"
                                        )
                                    }
                                )
                                IconButton(
                                    onClick = {
                                        currentRecipe = filteredItems.random()
                                        navController.navigate(
                                            RecipeAppScreen.Details.name.plus(
                                                "/${currentRecipe.title}"
                                            )
                                        )
                                    },
                                    content = {
                                        Icon(
                                            painter = painterResource(Res.drawable.shuffle),
                                            contentDescription = "random"
                                        )
                                    }
                                )
                                IconButton(
                                    onClick = {onThemeToggle.invoke()},
                                    content = {
                                        Icon(
                                            painter = painterResource(
                                                when(theme) {
                                                    Theme.Auto -> Res.drawable.auto_mode
                                                    Theme.Light -> Res.drawable.light_mode
                                                    Theme.Dark -> Res.drawable.dark_mode
                                                }
                                            ),
                                            contentDescription = "theme"
                                        )
                                    }
                                )
                            }
                        }
                }


                SharedTransitionLayout {
                    NavHost(
                        navController = navController,
                        startDestination = "Home",
                        modifier = Modifier.fillMaxSize()
                    ) {
                        composable(route = "Home")
                        {
                            searchBar = false
                            HomeScreen(
                                onClose = onClose,
                                onGo = {
                                    if (DisplayManager.device == Device.Web)
                                        navController.navigate(RecipeAppScreen.List.name)
                                    else
                                        navController.navigate("BookShelf")
                                },
                                onSurprise = {
                                    currentRecipe = filteredItems.random()
                                    navController.navigate(RecipeAppScreen.Details.name.plus("/${currentRecipe.title}"))
                                },
                            )
                        }
                        composable(route = "BookShelf")
                        {
                            val shelfTags = mutableMapOf<String, TagType>()

                            for (recipe in getRecipeList()) {
                                for (tag in recipe.tags) {
                                    if (!tags.contains(tag.key))
                                        shelfTags[tag.key] = tag.value
                                }
                            }

                            searchBar = false
                            BookShelf(
                                tags = shelfTags,
                                onClick = { lockedTag ->

                                    tags.clear()
                                    search = ""

                                    book = lockedTag
                                    if (book == "SURPRISE") {
                                        currentRecipe = filteredItems.random()
                                        navController.navigate(RecipeAppScreen.Details.name.plus("/${currentRecipe.title}"))
                                    } else {
                                        navController.navigate(RecipeAppScreen.List.name)
                                    }
                                }
                            )
                        }
                        composable(route = RecipeAppScreen.List.name) {
                            searchBar = true

                            RecipesListScreen(
                                items = filteredItems,
                                animate = animate,
                                onClick = { recipe ->
                                    currentRecipe = recipe
                                    navController.navigate(RecipeAppScreen.Details.name.plus("/${recipe.title}"))
                                })
                        }
                        composable(
                            route = RecipeAppScreen.Details.name.plus("/{recipe}"),
                            arguments = listOf(navArgument("recipe") {
                                type = NavType.StringType
                            })
                        ) { args ->
                            searchBar = false
                            getRecipeList().find { it.title == args.arguments?.getString("recipe") }
                                ?.let { it1 ->
                                    RecipeDetails(
                                        navController = navController,
                                        recipe = it1,
                                        animate = animate,
                                        multipleRecipes = filteredItems.size > 1,
                                        goForward = {

                                            val index = filteredItems.indexOf(it1)
                                            val recipe =
                                                if (index < filteredItems.size - 1) filteredItems[index + 1] else filteredItems.first()

                                            navController.navigate(
                                                RecipeAppScreen.Details.name.plus(
                                                    "/${recipe.title}"
                                                )
                                            )
                                        })
                                }
                        }

                        composable(
                            route = "Settings"
                        ) {
                            searchBar = false
                            SettingsMenu(
                                andOr = returnAnyMatch,
                                theme = theme,
                                sortBy = sortBy,
                                ttt = tttBy,
                                helpUs = helpUs,
                                descending = descending,
                                animate = animate,
                                onAndOr = {
                                    returnAnyMatch = !returnAnyMatch
                                    settings.putBoolean(AON_KEY, returnAnyMatch)
                                },
                                onSortToggle = { onSortToggle.invoke() },
                                onDesc = {
                                    descending = !descending
                                    settings.putBoolean(DESC_KEY, descending)
                                },
                                onAnim = {
                                    animate = !animate
                                    settings.putBoolean(ANIM_KEY, animate)
                                },
                                onTheme = { onThemeToggle.invoke() },
                                onHelpToggle = { helpUs = !helpUs },
                                onTTTToggle = { onTTTToggle.invoke() },
                                onRandom = {},
                                onBack = { navController.popBackStack() },
                                onClear = {settings.clear()}
                            )
                        }
                    }
                }
            }
        }
    }
}